"""Helper functions to work with filenames, directories etc."""

import os
import os.path
import platform
import re
from os import sep

from . import strtools
from .log import LOG as log


def parse_path(path, prefix=""):
    r"""Parse a path into its components.

    If the path doesn't end with the pathsep, it is assumed being a file!
    No tests based on existing files are done, as this is supposed to also work
    on path strings that don't exist on the system running this code.

    The function accepts `java.io.File` objects (as retrieved by using ImageJ2's
    *Script Parameter* `#@ File`) for either of the parameters, so it is safe to
    use this in ImageJ Python scripts without additional measures.

    **WARNING**: when passing in **Windows paths** literally, make sure to
    declare them as **raw strings** using the `r""` notation, otherwise
    unexpected things might happen if the path contains sections that Python
    will interpret as escape sequences (e.g. `\n`, `\t`, `\u2324`, ...).

    Parameters
    ----------
    path : str or str-like
        The path to be parsed into components.
    prefix : str or str-like, optional
        An optional path component that will be prefixed to the given path using
        `os.path.join()`.

    Returns
    -------
    dict
        The parsed (and possibly combined) path split into its components, with
        the following keys:

        - `orig` : The full string as passed into this function (possibly
          combined with the prefix in case one was specified).
        - `full` : The same as `orig` with separators adjusted to the current
          platform.
        - `parent` : The parent folder of the selected file.
        - `path` : The same as `full`, up to (including) the last separator.
        - `dname` : The segment between the last two separators (directory).
        - `fname` : The segment after the last separator (filename).
        - `basename` : The filename without extension. Note that *OME-TIFF*
          files (having a suffix like `.ome.tif` or `.ome.tiff`) are treated as
          special case in the sense that the `.ome` part is also stripped from
          the basename and added to the `ext` key (see below).
        - `ext` : The filename extension, containing max 1 dot (included) with
          the special case of `.ome.tif` / `.ome.tiff` where 2 dots are
          contained to represent the full suffix.

    Examples
    --------
    POSIX-style path to a file with a suffix:

    >>> parse_path('/tmp/foo/file.suffix')
    {
        "dname": "foo",
        "ext": "",
        "fname": "file",
        "full": "/tmp/foo/file",
        "basename": "file",
        "orig": "/tmp/foo/file",
        "parent": "/tmp/",
        "path": "/tmp/foo/",
    }


    POSIX-style path to a directory:

    >>> parse_path('/tmp/foo/')
    {
        "dname": "foo",
        "ext": "",
        "fname": "",
        "full": "/tmp/foo/",
        "basename": "",
        "orig": "/tmp/foo/",
        "parent": "/tmp/",
        "path": "/tmp/foo/",
    }


    Windows-style path to a file:

    >>> parse_path(r'C:\Temp\new\file.ext')
    {
        "dname": "new",
        "ext": ".ext",
        "fname": "file.ext",
        "full": "C:/Temp/new/file.ext",
        "basename": "file",
        "orig": "C:\\Temp\\new\\file.ext",
        "parent": "C:/Temp",
        "path": "C:/Temp/new/",
    }


    Special treatment for *OME-TIFF* suffixes:

    >>> parse_path("/path/to/some/nice.OME.tIf")
    {
        "basename": "nice",
        "dname": "some",
        "ext": ".OME.tIf",
        "fname": "nice.OME.tIf",
        "full": "/path/to/some/nice.OME.tIf",
        "orig": "/path/to/some/nice.OME.tIf",
        "parent": "/path/to/",
        "path": "/path/to/some/",
    }
    """
    path = str(path)
    if prefix:
        # remove leading slash, otherwise join() will discard the first path:
        if path.startswith("/"):
            path = path[1:]
        path = os.path.join(str(prefix), path)
    parsed = {}
    parsed["orig"] = path
    path = path.replace("\\", sep)
    parsed["full"] = path
    folder = os.path.dirname(path)
    parsed["path"] = folder + sep
    parsed["parent"] = os.path.dirname(folder)
    parsed["fname"] = os.path.basename(path)
    parsed["dname"] = os.path.basename(os.path.dirname(parsed["path"]))
    base, ext = os.path.splitext(parsed["fname"])
    parsed["ext"] = ext
    parsed["basename"] = base
    if base.lower().endswith(".ome") and ext.lower().startswith(".tif"):
        parsed["basename"] = base[:-4]
        parsed["ext"] = base[-4:] + ext

    return parsed


def join2(path1, path2):
    r"""Join two paths into one, much like os.path.join().

    The main difference is that `join2()` takes exactly two arguments, but they
    can be non-str (as long as they're having a `__str__()` method), so this is
    safe to be used with stuff like `java.io.File` objects as retrieved when
    using ImageJ2's *Script Parameter* `#@ File`.

    In addition some sanitizing is done, e.g. in case one of the components is
    containing double backslashes (`\\`), they will be replaced by the current
    OS's path separator.

    Parameters
    ----------
    path1 : str or str-like
        The first component of the path to be joined.
    path2 : str or str-like
        The second component of the path to be joined.

    Returns
    -------
    str
    """
    return parse_path(path2, prefix=path1)["full"]


def jython_fiji_exists(path):
    """Work around problems with `os.path.exists()` in Jython 2.7 in Fiji.

    In current Fiji, the Jython implementation of os.path.exists(path) raises a
    java.lang.AbstractMethodError iff 'path' doesn't exist. This function
    catches the exception to allow normal usage of the exists() call.
    """
    try:
        return os.path.exists(path)
    except java.lang.AbstractMethodError:  # pragma: no cover
        return False


def listdir_matching(
    path, suffix, fullpath=False, sort=False, regex=False, recursive=False
):
    """Get a list of files in a directory matching a given suffix.

    Parameters
    ----------
    path : str
        The directory to scan for files.
    suffix : str
        The suffix to match filenames against.
    fullpath : bool, optional
        If set to True, the list returned by the function will contain the full
        paths to the matching files (the default is False, which will result in
        the file names only, without path).
    sort : bool, optional
        If set to True, the returned list will be sorted using
        `imcflibs.strtools.sort_alphanumerically()`.
    regex : bool, optional
        If set to True, uses the suffix-string as regular expression to match
        filenames. By default False.
    recursive : bool, optional
        If set to True, the directory tree will be traversed recursively and
        files in subfolders will be included. When `fullpath` is False and
        `recursive` is True the returned file names are relative to `path`
        (e.g. `subdir/file.ext`). Default is False.

    Returns
    -------
    list
        All file names in the directory matching the suffix (without path!).
    """
    matching_files = list()

    # Prepare regex if requested (case-insensitive)
    regex_compiled = None
    if regex:
        try:
            regex_compiled = re.compile(suffix, re.IGNORECASE)
        except re.error:
            # If provided regex is invalid, fall back to no matches
            return matching_files

    if recursive:
        # Walk directory tree and test each filename
        for dirpath, _, filenames in os.walk(path):
            for candidate in filenames:
                if not regex_compiled:
                    if candidate.lower().endswith(suffix.lower()):
                        if fullpath:
                            matching_files.append(os.path.join(dirpath, candidate))
                        else:
                            rel = os.path.relpath(
                                os.path.join(dirpath, candidate), path
                            )
                            matching_files.append(rel)
                else:
                    if regex_compiled.match(candidate):
                        if fullpath:
                            matching_files.append(os.path.join(dirpath, candidate))
                        else:
                            rel = os.path.relpath(
                                os.path.join(dirpath, candidate), path
                            )
                            matching_files.append(rel)
    else:
        # Non-recursive: only list entries in the given directory
        for candidate in os.listdir(path):
            if not regex_compiled:
                if candidate.lower().endswith(suffix.lower()):
                    if fullpath:
                        matching_files.append(os.path.join(path, candidate))
                    else:
                        matching_files.append(candidate)
            else:
                if regex_compiled.match(candidate):
                    if fullpath:
                        matching_files.append(os.path.join(path, candidate))
                    else:
                        matching_files.append(candidate)

    if sort:
        matching_files = strtools.sort_alphanumerically(matching_files)

    return matching_files


def image_basename(orig_name):
    """Return the file name component without suffix(es).

    Strip away the path and suffix of a given file name, doing a special
    treatment for the composite suffix ".ome.tif(f)" which will be fully
    stripped as well.

    Parameters
    ----------
    orig_name : str
        The original name, possibly containing paths and filename suffix.

    Examples
    --------
    >>> image_basename('/path/to/some_funny_image_file_01.png')
    'some_funny_image_file_01'

    >>> image_basename('some-more-complex-stack.ome.tif')
    'some-more-complex-stack'

    >>> image_basename('/tmp/FoObAr.OMe.tIf')
    'FoObAr'
    """
    return parse_path(orig_name)["basename"]


def gen_name_from_orig(path, orig_name, tag, suffix):
    """Derive a file name from a given input file, an optional tag and a suffix.

    Parameters
    ----------
    path : str or object that can be cast to a str
        The output path.
    orig_name : str or object that can be cast to a str
        The input file name, may contain arbitrary path components.
    tag : str
        An optional tag to be added at the end of the new file name, can be used
        to denote information like "-avg" for an average projection image.
    suffix : str
        The new file name suffix, which also sets the file format for BF.

    Returns
    -------
    out_file : str
        The newly generated file name with its full path.
    """
    name = os.path.join(path, image_basename(orig_name) + tag + suffix)
    return name


def derive_out_dir(in_dir, out_dir):
    """Derive `out_dir` from its own value and the value of `in_dir`.

    In case the supplied value of `out_dir` is one of '-' or 'NONE', the
    returned value will be set to `in_dir`. Otherwise the value of `out_dir`
    will be returned unchanged.

    Parameters
    ----------
    in_dir : str
        The full path to the input directory.
    out_dir : str
        Either the full path to an output directory or one of '-' or 'NONE'.

    Returns
    -------
    str
        The full path to the directory to be used for output and temp files.
    """
    if out_dir.upper() in ["-", "NONE"]:
        out_dir = in_dir
        log.info("No output directory given, using input dir [%s].", out_dir)
    else:
        log.info("Using directory [%s] for results and temp files.", out_dir)

    return out_dir


def find_dirs_containing_filetype(source, filetype):
    """Recursively list directories containing files with a given suffix.

    Parameters
    ----------
    source : str
        Path to base directory to start recursive search in.
    filetype : str
        Filetype (string pattern) that should be matched against filenames in
        the directories.

    Returns
    -------
    list(str)
        List of all dirs that contain files with the given suffix / filetype.
    """
    dirs_containing_filetype = []

    # walk recursively through all directories
    # list their paths and all files inside (=os.walk)
    for dirname, _, filenames in os.walk(source):
        # stop when encountering a directory that contains "filetype"
        # and store the directory path
        for filename in filenames:
            if filetype in filename:
                dirs_containing_filetype.append(dirname + "/")
                break

    return dirs_containing_filetype


def folder_size(source):
    """Get the total size of a given directory and its subdirectories.

    Parameters
    ----------
    source : str
        Directory for which the size should be determined.

    Returns
    -------
    int
        The total size of all files in the source dir and subdirs in bytes.
    """
    total_size = 0
    for dirpath, _, filenames in os.walk(source):
        for fname in filenames:
            fpath = os.path.join(dirpath, fname)
            # skip if it is symbolic link
            if not os.path.islink(fpath):
                total_size += os.path.getsize(fpath)

    return total_size


def create_directory(new_path):
    """Create a new directory at the specified path.

    This is a workaround for Python 2.7 where `os.makedirs()` is lacking
    the `exist_ok` parameter that is present in Python 3.2 and newer.

    Parameters
    ----------
    new_path : str
        Path where the new directory should be created.
    """

    if not os.path.exists(new_path):
        os.makedirs(new_path)


# pylint: disable-msg=C0103
#   we use the variable name 'exists' in its common spelling (lowercase), so
#   removing this workaround will be straightforward at a later point
if platform.python_implementation() == "Jython":  # pragma: no cover
    # pylint: disable-msg=F0401
    #   java.lang is only importable within Jython, pylint would complain
    import java.lang

    exists = jython_fiji_exists
else:
    exists = os.path.exists


def join_files_with_channel_suffix(files, nchannels):
    """Join filenames and append channel-suffixed copies.

    For each filename in `files`, return a list where original filenames
    appear first followed by copies with suffixes `_0` .. `_{n-2}`
    (inserted before the file extension). This is suitable for passing
    to Bioformats/Jython in ``show_list`` mode when each channel is stored
    as a separate file.

    Parameters
    ----------
    files : list or tuple
        List or tuple of filename strings.
    nchannels : int
        Number of channels (>=1). If ``nchannels`` is 1 no suffixed copies
        are added.

    Returns
    -------
    list of str
        Ordered list of filenames (originals then suffixed copies).
    """
    if not files:
        return ""
    try:
        x = range(int(nchannels) - 1)
    except Exception:
        x = [0]
    suff = "_" + str(x)
    out = []
    # Keep original order, then add suffixed copies
    for f in files:
        out.append(f)
    for i in x:
        suff = "_" + str(i)
        for f in files:
            base, ext = os.path.splitext(f)
            out.append(base + suff + ext)
    return out



"""Prevent namespace clashes / race conditions when running in Jython."""

import importlib

# Using `import io` will heavily depend on the state of the Python engine and
# may fail when being used as a library inside Jython (as `io` might be shadowed
# by the Java package with the same name then) with something like this:
## AttributeError: 'javapackage' object has no attribute 'BufferedIOBase'
io = importlib.import_module("io")

try:
    # Python 2: "file" is built-in
    file_types = file, io.IOBase
except NameError:
    # Python 3: "file" fully replaced with IOBase
    file_types = (io.IOBase,)



"""Functions to work on stitching datasets."""

from os.path import join

from java.lang.System import getProperty  # pylint: disable-msg=import-error

import micrometa  # pylint: disable-msg=import-error
import ij  # pylint: disable-msg=import-error

from .misc import show_status, show_progress, error_exit
from ..strtools import flatten
from ..log import LOG as log


def process_fluoview_project(infile):
    """Process a FluoView mosaic project file.

    Parameters
    ----------
    infile : str
        The full path to either an `MATL_Mosaic.log` or any `*.omp2info` file
        from a FluoView MATL acquisition.

    Returns
    -------
    micrometa.experiment.MosaicExperiment
        The mosaic object if parsing the project file has been successful.
    """
    if infile[-9:] == ".omp2info":
        mosaicclass = micrometa.fluoview.FluoView3kMosaic
    elif infile[-4:] == ".log":
        mosaicclass = micrometa.fluoview.FluoViewMosaic
    else:
        error_exit("Unsupported input file: %s" % infile)

    log.info("Parsing project file: [%s]", infile)
    ij.IJ.showStatus("Parsing mosaics...")

    mosaics = mosaicclass(infile, runparser=False)
    total = len(mosaics.mosaictrees)
    ij.IJ.showProgress(0.0)
    show_status("Parsed %s / %s mosaics" % (0, total))
    for i, subtree in enumerate(mosaics.mosaictrees):
        log.info("Parsing mosaic %s...", i + 1)
        try:
            mosaics.add_mosaic(subtree, i)
        except (ValueError, IOError) as err:
            log.warn("Skipping mosaic %s: %s", i, err)
        except RuntimeError as err:
            log.warn("Error parsing mosaic %s, SKIPPING: %s", i, err)
        show_progress(i, total)
        show_status("Parsed %s / %s mosaics" % (i + 1, total))
    show_progress(total, total)
    show_status("Parsed %i mosaics." % total)

    if not mosaics:
        error_exit("Couldn't find any (valid) mosaics in the project file!")
    log.info(mosaics.summarize())

    return mosaics


def gen_macro(mosaics, indir, outfile=None, opts=None):
    """Generate stitcher macro code and optionally save it in a file.

    Parameters
    ----------
    mosaics : micrometa.experiment.MosaicExperiment
        The mosaic object of the stitching experiment.
    indir : str
        The path to use as input directory *INSIDE* the macro.
    outfile : str (optional)
        The path to a file for saving the generated macro code.
    opts : dict (optional)
        A dict to be passed on to micrometa.imagej.gen_stitching_macro().

    Returns
    -------
    list(str)
        The generated macro code as a list of strings (one str per line).
    """
    templates = join(getProperty("fiji.dir"), "jars", "python-micrometa.jar")
    log.info("Using macro templates from [%s].", templates)
    log.info("Using [%s] as base directory.", indir)

    # set the default stitcher options
    stitcher_options = {
        "export_format": '".ids"',
        "split_z_slices": "false",
        "rotation_angle": 0,
        "stitch_regression": 0.3,
        "stitch_maxavg_ratio": 2.5,
        "stitch_abs_displace": 3.5,
        "compute": "false",
    }
    # merge explicit options, overriding the defaults from above if applicable:
    if opts:
        stitcher_options.update(opts)

    code = micrometa.imagej.gen_stitching_macro(
        name=mosaics.infile["dname"],
        path=indir,
        tplpfx="templates/imagej-macro/stitching",
        tplpath=templates,
        opts=stitcher_options,
    )

    log.debug("============= begin of generated macro code =============")
    log.debug(flatten(code))
    log.debug("============= end of generated  macro code =============")

    if outfile is not None:
        log.info("Writing stitching macro.")
        micrometa.imagej.write_stitching_macro(code, outfile)

    return code



# -*- coding: utf-8 -*-

"""Functions to work with ImageJ label images."""

from ij import IJ, ImagePlus, ImageStack, Prefs
from ij.plugin import Duplicator, ImageCalculator
from ij.plugin.filter import ThresholdToSelection
from ij.process import FloatProcessor, ImageProcessor
from inra.ijpb.label import LabelImages as li
from inra.ijpb.plugins import AnalyzeRegions
from mcib3d.image3d import ImageHandler, ImageLabeller


def label_image_to_roi_list(label_image, low_thresh=None):
    """Convert a label image to a list of ROIs.

    Parameters
    ----------
    label_image : ij.ImagePlus
        Label image to convert.
    low_thresh : int, optional
        Value under which the labels should be discarded, by default `None`.

    Returns
    -------
    roi_list : list(ij.gui.Roi)
        List of all the ROIs converted from the label image.
    """

    roi_list = []
    max_value = 0

    for slice in range(1, label_image.getNSlices() + 1):
        label_image_slice = Duplicator().run(label_image, 1, 1, slice, slice, 1, 1)

        image_processor = label_image_slice.getProcessor()
        pixels = image_processor.getFloatArray()

        existing_pixel_values = set()

        for x in range(0, image_processor.getWidth()):
            for y in range(0, image_processor.getHeight()):
                existing_pixel_values.add(pixels[x][y])

        # Converts data in case it's RGB image
        float_processor = FloatProcessor(
            image_processor.getWidth(), image_processor.getHeight()
        )
        float_processor.setFloatArray(pixels)
        img_float_copy = ImagePlus("FloatLabel", float_processor)

        # print(existing_pixel_values)
        for value in existing_pixel_values:
            if low_thresh is not None:
                if value < low_thresh:
                    continue
            elif value == 0:
                continue
            # print(value)
            float_processor.setThreshold(value, value, ImageProcessor.NO_LUT_UPDATE)
            roi = ThresholdToSelection.run(img_float_copy)
            roi.setName(str(value))
            roi.setPosition(slice)
            roi_list.append(roi)
            max_value = max(max_value, value)

    return roi_list, max_value


def cookie_cut_labels(label_image_ref, label_image_to_relate):
    """Relate label images, giving the same label to objects belonging together.

    ❗ NOTE: Won't work with touching labels ❗

    Given two label images, this function will create a new label image
    with the same labels as the reference image, but with the objects
    of the second image.

    Parameters
    ----------
    label_image_ref : ij.ImagePlus
        Reference to use for the labels.
    label_image_to_relate : ij.ImagePlus
        Image to change for the labels.

    Returns
    -------
    ij.ImagePlus
        New ImagePlus with modified labels matching the reference.
    """

    imp_dup = label_image_to_relate.duplicate()
    IJ.setRawThreshold(imp_dup, 1, 65535)
    Prefs.blackBackground = True
    IJ.run(imp_dup, "Convert to Mask", "")
    IJ.run(imp_dup, "Divide...", "value=255")
    return ImageCalculator.run(label_image_ref, imp_dup, "Multiply create")


def relate_label_images(outer_label_imp, inner_label_imp):
    """Relate label images, giving the same label to objects belonging together.

    Given two label images, this function will create a new label image with the
    same labels as the reference image, but with the objects of the second image
    using the 3D Association plugin from the 3DImageJSuite.

    Parameters
    ----------
    outer_label_imp : ij.ImagePlus
        The outer label image.
    inner_label_imp : ij.ImagePlus
        The inner label image.

    Returns
    -------
    related_inner_imp : ij.ImagePlus
        The related inner label image.

    Notes
    -----
    Unlike `cookie_cut_labels`, this should work with touching labels by using
    MereoTopology algorithms.
    """

    outer_label_imp.show()
    inner_label_imp.show()

    outer_title = outer_label_imp.getTitle()
    inner_title = inner_label_imp.getTitle()

    IJ.run(
        "3D Association",
        "image_a="
        + outer_title
        + " "
        + "image_b="
        + inner_title
        + " "
        + "method=Colocalisation min=1 max=0.000",
    )

    related_inner_imp = IJ.getImage()

    outer_label_imp.hide()
    inner_label_imp.hide()
    related_inner_imp.hide()

    return related_inner_imp


def filter_objects(label_image, table, string, min_val, max_val):
    """Filter labels based on specific min and max values.

    Parameters
    ----------
    label_image : ij.ImagePlus
        Label image on which to filter.
    table : ResultsTable
        ResultsTable containing all the measurements on which to filter.
    string : str
        Measurement name on which to filter, e.g. `Area`, `Mean Intensity` etc.
    min_val : float
        Minimum value to keep.
    max_val : float
        Maximum value to keep

    Returns
    -------
    ij.ImagePlus
        Label image containing only the remaining labels.
    """

    keep_label_id = []
    for row in range(table.size()):
        current_value = table.getValue(string, row)
        if current_value >= min_val and current_value <= max_val:
            keep_label_id.append(int(table.getLabel(row)))

    return li.keepLabels(label_image, keep_label_id)


def measure_objects_size_shape_2d(label_image):
    """Measure the shapes of the different labels.

    Parameters
    ----------
    label_image : ij.ImagePlus
        Label image on which to get the shapes.

    Returns
    -------
    ResultsTable
        ResultsTable with the shape measurements.
    """
    regions = AnalyzeRegions()
    return regions.process(label_image)


def binary_to_label(imp, title, min_thresh=1, min_vol=None, max_vol=None):
    """Segment a binary image to get a label image (2D/3D).

    Works on: 2D and 3D binary data.

    Parameters
    ----------
    imp : ij.ImagePlus
        Binary 3D stack or 2D image.
    title : str
        Title of the new image.
    min_thresh : int, optional
        Threshold to do segmentation, also allows for label filtering, by default 1.
    min_vol : float, optional
        Volume under which to exclude objects, by default None.
    max_vol : float, optional
        Volume above which to exclude objects, by default None.

    Returns
    -------
    ij.ImagePlus
        Segmented labeled ImagePlus.
    """
    # Get the calibration of the input ImagePlus
    cal = imp.getCalibration()

    # Wrap the ImagePlus in an ImageHandler
    img = ImageHandler.wrap(imp)

    # Threshold the image using the specified threshold value
    img = img.threshold(min_thresh, False, False)

    # Create an ImageLabeller instance
    labeler = ImageLabeller()

    # Set the minimum size for labeling if provided
    if min_vol:
        labeler.setMinSizeCalibrated(min_vol)

    # Set the maximum size for labeling if provided
    if max_vol:
        labeler.setMinSizeCalibrated(max_vol)

    # Get the labeled image
    seg = labeler.getLabels(img)

    # Set the scale of the labeled image
    seg.setScale(cal.pixelWidth, cal.pixelDepth, cal.getUnits())

    # Set the title of the labeled image
    seg.setTitle(title)

    # Return the segmented labeled ImagePlus
    return seg.getImagePlus()


def dilate_labels_2d(imp, dilation_radius):
    """Dilate each label in the given ImagePlus using the specified dilation radius.

    This method will use a 2D dilation to be applied to each slice of the ImagePlus
    and return a new stack.

    Parameters
    ----------
    imp : ij.ImagePlus
        Input ImagePlus with the labels to dilate
    dilation_radius : int
        Number of pixels to dilate each label

    Returns
    -------
    ij.ImagePlus
        New ImagePlus with the dilated labels
    """

    # Create a list of the dilated labels
    dilated_labels_list = []

    # Iterate over each slice of the input ImagePlus
    for i in range(1, imp.getNSlices() + 1):
        # Duplicate the current slice
        current_imp = Duplicator().run(imp, 1, 1, i, imp.getNSlices(), 1, 1)

        # Perform a dilation of the labels in the current slice
        dilated_labels_imp = li.dilateLabels(current_imp, dilation_radius)

        # Append the dilated labels to the list
        dilated_labels_list.append(dilated_labels_imp)

    # Create a new ImagePlus with the dilated labels
    dilated_labels_imp = ImagePlus(
        "Dilated labels", ImageStack().create(dilated_labels_list)
    )

    return dilated_labels_imp



"""Python convenience from and for the IMCF (Imaging Core Facility).

.. include:: ../../README.md

.. include:: ../../TESTING.md

.. include:: ../../CHANGELOG.md
"""

__version__ = "2.0.0.a3"

from . import iotools
from . import log
from . import pathtools
from . import strtools

# check if we're running in Jython, then also import the 'imagej' submodule:
import platform as _python_platform

if _python_platform.python_implementation() == "Jython":  # pragma: no cover
    from . import imagej
del _python_platform



"""Convenience wrappers around CLIJ and GPU accelerated functions."""

from ij.plugin import Duplicator, RGBStackMerge  # pylint: disable-msg=import-error


def erode_labels(clij2_instance, label_image, erosion_radius, channel=None):
    """Erode labels using GPU acceleration.

    Parameters
    ----------
    clij2_instance : net.haesleinhuepf.clij.CLIJ
        Instance of CLIJ to communicate with the GPU.
    label_image : ij.ImagePlus
        Label image to be eroded.
    erosion_radius : int
        Radius for erosion.
    channel : int, optional
        Specific channel to apply erosion.

    Returns
    -------
    ij.ImagePlus
        Label image with eroded labels.
    """

    list_of_images = []
    channel_list = [channel] if channel else range(1, label_image.getNChannels() + 1)

    for channel in channel_list:
        current_channel = Duplicator().run(
            label_image,
            channel,
            channel,
            1,
            label_image.getNSlices(),
            1,
            label_image.getNFrames(),
        )

        # clij2 = CLIJ2.getInstance()
        src = clij2_instance.push(current_channel)
        dst = clij2_instance.create(src)
        mask = clij2_instance.create(src)

        clij2_instance.erodeLabels(src, dst, erosion_radius, False)
        clij2_instance.mask(src, dst, mask)

        list_of_images.append(clij2_instance.pull(mask))

    if len(list_of_images) > 1:
        return RGBStackMerge().mergeChannels(list_of_images, False)
    else:
        return list_of_images[0]


def dilate_labels(clij2_instance, label_image, dilation_radius, channel=None):
    """Dilate labels using GPU acceleration.

    Parameters
    ----------
    clij2_instance : net.haesleinhuepf.clij.CLIJ
        Instance of CLIJ to communicate with the GPU.
    label_image : ij.ImagePlus
        Label Image to be dilated.
    dilation_radius : int
        Radius for dilation.
    channel : int, optional
        Specific channel to apply dilation.

    Returns
    -------
    ij.ImagePlus
        Label image with dilated labels.
    """

    list_of_images = []
    channel_list = [channel] if channel else range(1, label_image.getNChannels() + 1)

    for channel in channel_list:
        print(channel)

        current_channel = Duplicator().run(
            label_image,
            channel,
            channel,
            1,
            label_image.getNSlices(),
            1,
            label_image.getNFrames(),
        )

        # clij2 = CLIJ2.getInstance()
        src = clij2_instance.push(current_channel)
        dst = clij2_instance.create(src)
        mask = clij2_instance.create(src)

        clij2_instance.dilateLabels(src, dst, dilation_radius)
        # clij2_instance.mask(src, dst, mask) # Seems that for dilation no masking is needed ?

        list_of_images.append(clij2_instance.pull(dst))
        # clij2_instance.pull(mask).show()

        clij2_instance.clear()

    if len(list_of_images) > 1:
        return RGBStackMerge().mergeChannels(list_of_images, False)
    else:
        return list_of_images[0]


def merge_labels(clij2_instance, label_image, channel=None):
    """Merge touching labels using GPU acceleration.

    Parameters
    ----------
    clij2_instance : net.haesleinhuepf.clij.CLIJ
        Instance of CLIJ to communicate with the GPU.
    label_image : ij.ImagePlus
        Label image with touching labels.
    channel : int, optional
        Specific channel to apply label merging.

    Returns
    -------
    ij.ImagePlus
        New ImagePlus with merged labels.
    """

    list_of_images = []

    if channel is None:
        channel_list = range(1, label_image.getNChannels() + 1)
    else:
        channel_list = [channel]

    for channel in channel_list:
        current_channel = Duplicator().run(
            label_image,
            channel,
            channel,
            1,
            label_image.getNSlices(),
            1,
            label_image.getNFrames(),
        )

        src = clij2_instance.push(current_channel)
        dst = clij2_instance.create(src)

        clij2_instance.mergeTouchingLabels(src, dst)
        list_of_images.append(clij2_instance.pull(dst))

        clij2_instance.clear()

    if len(list_of_images) > 1:
        return RGBStackMerge().mergeChannels(list_of_images, False)

    return list_of_images[0]



"""Common logging singleton for the package.

Example
-------
>>> from log import LOG as log

From there on a logger is available for usage with e.g. log.warn(), even if the
import statement from above happens in multiple places across modules, it will
always use the same logger instance (that "singleton" functionality is built
into the logging module, we just do the setup here). This can easily be checked
by looking at the log handlers in the different modules.

The logging levels, in increasing order of importance, are:

10 DEBUG
20 INFO
30 WARN
40 ERROR
50 CRITICAL
"""


import logging


LOG = logging.getLogger(__name__)


def enable_console_logging():
    """Add a stream handler logging to the console.

    Returns
    -------
    logging.StreamHandler
    """
    stream_handler = logging.StreamHandler()
    LOG.addHandler(stream_handler)
    return stream_handler


def enable_file_logging(fname, mode="a"):
    """Add a logging handler writing to a file.

    Returns
    -------
    logging.FileHandler
    """
    file_handler = logging.FileHandler(fname, mode=mode)
    LOG.addHandler(file_handler)
    return file_handler


def set_loglevel(verbosity):
    """Calculate the default loglevel and set it accordingly.

    This is a convenience function that wraps the calculation and setting of
    the logging level. The way our "log" module is currently built (as a
    singleton), there is no obvious better way to have this somewhere else.

    Example
    -------
    This will set the loglevel to DEBUG:
    >>> set_loglevel(2)

    This will set the loglevel to INFO:
    >>> set_loglevel(1)

    To set the verbosity level when you're e.g. using argparse to count the
    number of occurences of '-v' switches on the commandline into a variable
    'verbosity', this code can be used:
    >>> log.set_loglevel(args.verbosity)
    """
    # default loglevel is 30 while 20 and 10 show more details
    loglevel = (3 - verbosity) * 10
    LOG.setLevel(loglevel)



"""Functions for creating projections."""

from ij.plugin import ZProjector  # pylint: disable-msg=E0401

from .bioformats import export_using_orig_name  # pylint: disable-msg=E0401
from ..log import LOG as log

from net.imagej.axis import Axes
from net.imagej.ops import Ops
from ij import ImagePlus, IJ
from net.imagej import Dataset


def average(imp):
    """Create an average intensity Z projection.

    Parameters
    ----------
    imp : ij.ImagePlus
        The input stack to be projected.

    Returns
    -------
    ij.ImagePlus
        The result of the projection.
    """
    if imp.getDimensions()[3] < 2:
        log.warn("ImagePlus is not a z-stack, not creating a projection!")
        return imp

    log.debug("Creating average Z projection...")
    proj = ZProjector.run(imp, "avg")
    return proj


def maximum(imp):
    """Create a maximum intensity Z projection.

    Parameters
    ----------
    imp : ij.ImagePlus
        The input stack to be projected.

    Returns
    -------
    ij.ImagePlus
        The result of the projection.
    """
    if imp.getDimensions()[3] < 2:
        log.warn("ImagePlus is not a z-stack, not creating a projection!")
        return imp

    log.debug("Creating maximum intensity Z projection...")
    proj = ZProjector.run(imp, "max")
    return proj


def create_and_save(imp, projections, path, filename, export_format):
    """Create one or more projections and export (save) them.

    Parameters
    ----------
    imp : ij.ImagePlus
        The image stack to create the projections from.
    projections : list(str)
        A list of projection types to be done, valid options are 'Average',
        'Maximum' and 'Sum'.
    path : str
        The path to store the results in. Existing files will be overwritten.
    filename : str
        The original file name to derive the results name from.
    export_format : str
        The suffix to be given to Bio-Formats, determining the storage format.

    Returns
    -------
    bool
        True in case projections were created, False otherwise (e.g. if the
        given ImagePlus is not a Z-stack).
    """
    if not projections:
        log.debug("No projection type requested, skipping...")
        return False

    if imp.getDimensions()[3] < 2:
        log.error("ImagePlus is not a z-stack, not creating any projections!")
        return False

    command = {
        "Average": "avg",
        "Maximum": "max",
        "Sum": "sum",
    }
    for projection in projections:
        log.debug("Creating '%s' projection...", projection)
        proj = ZProjector.run(imp, command[projection])
        export_using_orig_name(
            proj,
            path,
            filename,
            "-%s" % command[projection],
            export_format,
            overwrite=True,
        )
        proj.close()

    return True


def project_stack(imp, projected_dimension, projection_type, ops, ds, cs):
    """Project along a defined axis using the given projection type.

    Parameters
    ----------
    imp : ImagePlus
        The input image to be projected.
    projected_dimension : str
        The dimension along which to project the data. Must be one of {"X", "Y", "Z",
        "TIME", "CHANNEL"}.
    projection_type : str
        The type of projection to perform. Must be one of {"Max", "Mean", "Median",
        "Min", "StdDev", "Sum"}.
    ops : OpService
        The service used to access image processing operations. Use e.g. from script
        parameter: `#@ OpService ops`
    ds : DatasetService
        The service used to create new datasets. Use e.g. from script parameter:
        `#@ DatasetService ds`
    cs : ConvertService
        The service used to convert between formats. Use e.g. from script parameter:
        `#@ ConvertService cs`

    Returns
    -------
    ImagePlus
        The resulting projected image as an ImagePlus object.

    Raises
    ------
    Exception
        If the specified dimension is not found or if the dimension has only one frame.
    """
    bit_depth = imp.getBitDepth()
    data = cs.convert(imp, Dataset)
    # Select which dimension to project
    dim = data.dimensionIndex(getattr(Axes, projected_dimension))
    if dim == -1:
        raise Exception("%s dimension not found." % projected_dimension)
    if data.dimension(dim) < 2:
        raise Exception("%s dimension has only one frame." % projected_dimension)

    # Write the output dimensions
    new_dimensions = [
        data.dimension(d) for d in range(0, data.numDimensions()) if d != dim
    ]

    # Create the output image
    projected = ops.create().img(new_dimensions)

    # Create the op and run it
    proj_op = ops.op(getattr(Ops.Stats, projection_type), data)
    ops.transform().project(projected, data, proj_op, dim)

    # Create the output Dataset and convert to ImagePlus
    output = ds.create(projected)
    output_imp = cs.convert(output, ImagePlus)
    output_imp = output_imp.duplicate()
    output_imp.setTitle("%s %s projection" % (projected_dimension, projection_type))
    IJ.run(output_imp, "Enhance Contrast", "saturated=0.35")

    # Rescale bit depth if possible
    if projection_type in ["Max", "Min", "Median"]:
        IJ.run("Conversions...", " ")
        if bit_depth in [8, 16]:
            IJ.run(output_imp, str(bit_depth) + "-bit", "")
        if bit_depth == 12:
            IJ.run(output_imp, "16-bit", "")

        IJ.run("Conversions...", "scale")

    return output_imp



"""ImageJ processing utilities for filtering and thresholding images.

This module provides functions to apply various image processing operations
using ImageJ, including filters, background subtraction, and thresholding.
"""

from ij import IJ

from ..log import LOG as log


def apply_filter(imp, filter_method, filter_radius, do_3d=False):
    """Make a specific filter followed by a threshold method of choice.

    Parameters
    ----------
    imp : ImagePlus
        Input ImagePlus to filter and threshold
    filter_method : str
        Name of the filter method to use. Must be one of:
            - Median
            - Mean
            - Gaussian Blur
            - Minimum
            - Maximum
    filter_radius : int
        Radius of the filter filter to use
    do_3d : bool, optional
        If set to True, will do a 3D filtering, by default False


    Returns
    -------
    ij.ImagePlus
        Filtered ImagePlus
    """
    log.info("Applying filter %s with radius %d" % (filter_method, filter_radius))

    if filter_method not in [
        "Median",
        "Mean",
        "Gaussian Blur",
        "Minimum",
        "Maximum",
    ]:
        raise ValueError(
            "filter_method must be one of: Median, Mean, Gaussian Blur, Minimum, Maximum"
        )

    if do_3d:
        filter = filter_method + " 3D..."
    else:
        filter = filter_method + "..."

    options = (
        "sigma="
        if filter_method == "Gaussian Blur"
        else "radius=" + str(filter_radius) + " stack"
    )

    log.debug("Filter: <%s> with options <%s>" % (filter, options))

    imageplus = imp.duplicate()
    IJ.run(imageplus, filter, options)

    return imageplus


def apply_rollingball_bg_subtraction(imp, rolling_ball_radius, do_3d=False):
    """Perform background subtraction using a rolling ball method.

    Parameters
    ----------
    imp : ij.ImagePlus
        Input ImagePlus to filter and threshold
    rolling_ball_radius : int
        Radius of the rolling ball filter to use
    do_3d : bool, optional
        If set to True, will do a 3D filtering, by default False

    Returns
    -------
    ij.ImagePlus
        Filtered ImagePlus
    """
    log.info("Applying rolling ball with radius %d" % rolling_ball_radius)

    options = "rolling=" + str(rolling_ball_radius) + " stack" if do_3d else ""

    log.debug("Background subtraction options: %s" % options)

    imageplus = imp.duplicate()
    IJ.run(imageplus, "Substract Background...", options)

    return imageplus


def apply_threshold(imp, threshold_method, do_3d=True):
    """Apply a threshold method to the input ImagePlus.

    Parameters
    ----------
    imp : ij.ImagePlus
        Input ImagePlus to filter and threshold
    threshold_method : str
        Name of the threshold method to use
    do_3d : bool, optional
        If set to True, the automatic threshold will be done on a 3D stack, by default True

    Returns
    -------
    ij.ImagePlus
        Thresholded ImagePlus
    """

    log.info("Applying threshold method %s" % threshold_method)

    imageplus = imp.duplicate()

    auto_threshold_options = (
        threshold_method + " " + "dark" + " " + "stack" if do_3d else ""
    )

    log.debug("Auto threshold options: %s" % auto_threshold_options)

    IJ.setAutoThreshold(imageplus, auto_threshold_options)

    convert_to_binary_options = (
        "method=" + threshold_method + " " + "background=Dark" + " " + "black"
    )

    log.debug("Convert to binary options: %s" % convert_to_binary_options)

    IJ.run(imageplus, "Convert to Mask", convert_to_binary_options)

    return imageplus



"""Functions to work on shading correction / model generation."""

import os

import ij  # pylint: disable-msg=import-error
from ij import IJ
from ij.plugin import ImageCalculator
from ij.process import StackStatistics
from ..imagej import bioformats  # pylint: disable-msg=no-name-in-module
from ..imagej import misc, projections
from ..log import LOG as log
from ..pathtools import gen_name_from_orig, listdir_matching


def apply_model(imps, model, merge=True):
    """Apply a given shading model to a list of images / stacks.

    The model is supposed to be a normalized 32-bit floating point 2D image that
    will be used as a divisor to the slices of all ImagePlus objects given.

    WARNING: the operation happens in-place, i.e. the original "imps" images
    will be modified!

    Parameters
    ----------
    imps : list(ij.ImagePlus)
        A list of ImagePlus objects (e.g. separate channels of a multi-channel
        stack image) that should be corrected for shading artefacts.
    model : ij.ImagePlus
        A 2D image with 32-bit float values normalized to 1.0 (i.e. no pixels
        with higher values) to be used for dividing the input images to correct
        for shading.
    merge : bool, optional
        Whether or not to combine the resulting ImagePlus objects into a single
        multi-channel stack (default=True).

    Returns
    -------
    ij.ImagePlus or list(ij.ImagePlus)
        The merged ImagePlus with all channels, or the original list of stacks
        with the shading-corrected image planes.
    """
    log.debug("Applying shading correction...")
    calc = ij.plugin.ImageCalculator()
    for i, stack in enumerate(imps):
        log.debug("Processing channel %i...", i)
        calc.run("Divide stack", stack, model)

    if not merge:
        return imps

    log.debug("Merging shading-corrected channels...")
    merger = ij.plugin.RGBStackMerge()
    merged_imp = merger.mergeChannels(imps, False)
    return merged_imp


def correct_and_project(filename, path, model, proj, fmt):
    """Apply a shading correction to an image and create a projection.

    In case the target file for the shading corrected image already exists,
    nothing is done - neither the shading correction is re-created nor any
    projections will be done (independent on whether the latter one already
    exist or not).

    Parameters
    ----------
    filename : str
        The full path to a multi-channel image stack.
    path : str
        The full path to a directory for storing the results. Will be created in
        case it doesn't exist yet. Existing files will be overwritten.
    model : ij.ImagePlus or None
        A 32-bit floating point image to be used as the shading model. If model
        is None, no shading correction will be applied.
    proj : str
        A string describing the projections to be created. Use 'None' for not
        creating any projections, 'ALL' to do all supported ones.
    fmt : str
        The file format suffix to be used for the results and projections, e.g.
        '.ics' for ICS2 etc. See the Bio-Formats specification for details.

    Returns
    -------
    (bool, bool)
        A tuple of booleans indicating whether a shading correction has been
        applied and whether projections were created. The latter depends on
        both, the requested projections as well as the image type (e.g. it can
        be False even if projections were requested, but the image)
    """
    target = gen_name_from_orig(path, filename, "", fmt)
    if os.path.exists(target):
        log.info("Found shading corrected file, not re-creating: %s", target)
        return False, False

    if not os.path.exists(path):
        os.makedirs(path)

    imps = bioformats.import_image(filename, split_c=True)
    ret_corr = False
    if model is not None:
        log.debug("Applying shading correction on [%s]...", filename)
        imp = apply_model(imps, model)
        bioformats.export_using_orig_name(imp, path, filename, "", fmt, True)
        # imps needs to be updated with the new (=merged) stack:
        imps = [imp]
        ret_corr = True

    if proj == "None":
        projs = []
    elif proj == "ALL":
        projs = ["Average", "Maximum"]
    else:
        projs = [proj]
    for imp in imps:
        ret_proj = projections.create_and_save(imp, projs, path, filename, fmt)
        imp.close()

    log.debug("Done processing [%s].", os.path.basename(filename))
    return ret_corr, ret_proj


def process_folder(path, suffix, outpath, model_file, fmt):
    """Run shading correction and projections on an entire folder.

    Parameters
    ----------
    path : str
        The input folder to be scanned for images to be processed.
    suffix : str
        The file name suffix of the files to be processed.
    outpath : str
        The output folder where results will be stored. Existing files will be
        overwritten.
    model_file : str
        The full path to a normalized 32-bit shading model image. If set to '-'
        or 'NONE', no shading correction will be applied, i.e. only the
        projection step will have an effect.
    fmt : str
        The file format suffix for storing the results.
    """
    matching_files = listdir_matching(path, suffix, fullpath=True)
    process_files(matching_files, outpath, model_file, fmt)


def process_files(files, outpath, model_file, fmt):
    """Run shading correction and projections on a list of files.

    Parameters
    ----------
    files : list(str)
        The files to be processed, as a list of strings with the full path.
    outpath : str
        The output folder where results will be stored. Existing files will be
        overwritten.
    model_file : str
        The full path to a normalized 32-bit shading model image. If set to '-'
        or 'NONE', no shading correction will be applied, i.e. only the
        projection step will have an effect.
    fmt : str
        The file format suffix for storing the results.
    """
    log.info("Running shading correction and projections on %s files...", len(files))

    if model_file.upper() in ["-", "NONE"]:
        model = None
    else:
        model = ij.IJ.openImage(model_file)
        # the model needs to be shown, otherwise the IJ.run() call ignores it
        try:
            model.show()
            canvas = model.getCanvas()
            for _ in range(5):
                # we have to show it, but we can make it smaller:
                canvas.zoomOut(100, 100)
        except AttributeError:
            misc.error_exit("Opening shading model [%s] failed!" % model_file)

    for in_file in files:
        correct_and_project(in_file, outpath, model, "ALL", fmt)

    if model:
        model.close()


def simple_flatfield_correction(imp, sigma=20.0):
    """Perform a simple flatfield correction to a given ImagePlus stack.

    Parameters
    ----------
    imp : ij.ImagePlus
        The input stack to be projected.
    sigma: float, optional
        The sigma value for the Gaussian blur, default=20.0.

    Returns
    -------
    ij.ImagePlus
        The 32-bit image resulting from the flatfield correction.
    """
    flatfield = imp.duplicate()
    sigma_str = "sigma=" + str(sigma)

    IJ.run(flatfield, "Gaussian Blur...", sigma_str)
    stats = StackStatistics(flatfield)

    # Normalize image to the highest value of original (requires 32-bit image)
    IJ.run(flatfield, "32-bit", "")
    IJ.run(flatfield, "Divide...", "value=" + str(stats.max))
    ic = ImageCalculator()
    flatfield_corrected = ic.run("Divide create", imp, flatfield)

    return flatfield_corrected



"""Internal wrapper module to import from the (Java) loci package.

This module exists to work around the issue that importing some of the classes
from the Java `loci` package would require syntax that is considered invalid by
any C-Python parser (but is still valid and working in Jython) and hence will
break usage of Black, Pylint, and similar.

By aggregating those "special" imports into this (private) submodule we can
properly work around that issue by providing "dummy" objects for C-Python and
importing the actual modules / classes when running within Jython. To avoid the
invalid syntax issue (which would still prevent C-Python-based tools like black
and pdoc to run) those parts are done via `importlib` calls.

Other loci-related imports (i.e. those without problematic syntax) are placed in
here simply for consistency reasons (to have everything in the same place).
"""

#
### *** WARNING *** ### *** WARNING *** ### *** WARNING *** ### *** WARNING ***
#
# Whenever an import is added here, make sure to also update the corresponding
# part in `imcf-fiji-mocks`: https://github.com/imcf/imcf-fiji-mocks/
#
### *** WARNING *** ### *** WARNING *** ### *** WARNING *** ### *** WARNING ***
#

# unproblematic imports kept here for consistency
from loci.plugins import BF
from loci.common import Region
from loci.formats import ImageReader, Memoizer, MetadataTools

# dummy objects to prevent failing imports in a non-ImageJ / Jython context:
ImporterOptions = None
ZeissCZIReader = None
DefaultMetadataOptions = None
MetadataLevel = None
DynamicMetadataOptions = None
MetadataOptions = None

# perform the actual imports when running under Jython using `importlib` calls:
import platform as _python_platform

if _python_platform.python_implementation() == "Jython":  # pragma: no cover
    import importlib

    _loci_plugins_in = importlib.import_module("loci.plugins.in")
    ImporterOptions = _loci_plugins_in.ImporterOptions

    _loci_formats_in = importlib.import_module("loci.formats.in")
    ZeissCZIReader = _loci_formats_in.ZeissCZIReader
    DefaultMetadataOptions = _loci_formats_in.DefaultMetadataOptions
    MetadataLevel = _loci_formats_in.MetadataLevel
    DynamicMetadataOptions = _loci_formats_in.DynamicMetadataOptions
    MetadataOptions = _loci_formats_in.MetadataOptions
del _python_platform



"""Functions to work with ImageJ preferences."""

from ij import Prefs, IJ  # pylint: disable-msg=E0401


def debug_mode():
    """Check if the 'imcf.debugging' setting is enabled.

    This is a workaround for a Jython issue in ImageJ with values that are
    stored in the "IJ_Prefs.txt" file being cast to the wrong types and / or
    values in Python. Calling Prefs.get() using a (Python) boolean as the
    second parameter always leads to the return value '0.0' (Python type float),
    no matter what is actually stored in the preferences. Doing the same in e.g.
    Groovy behaves correctly.

    Calling Prefs.get() as below with the second parameter being a string and
    subsequently checking the string value leads to the expected result.
    """
    debug = Prefs.get("imcf.debugging", "false")
    return debug == "true"


def set_default_ij_options():
    """Configure ImageJ default options for consistency.

    Will set the following options to ensure consistent behavior independent of
    how ImageJ is configured on a specific machine.

    - Ensure ImageJ appearance settings are the default values.
    - Set foreground color to white and background to black.
    - Set default file saving format to .txt files.
    - Ensure intensities are being scaled when converting between bit depths.
    - Options on binary images:
        - Set background to black.
        - Enable padding to prevent eroding from image edges.
        - Enforce defaults on iterations and count for *erosion*, *dilation*,
          *opening* and *closing* operations.

    References
    ----------
    The ImageJ User Guide is providing detailed explanations of the options
    configured by this function:

    - [Edit > Options > Appearance][ijo_app]
    - [Edit > Options > Colors][ijo_col]
    - [Edit > Options > Conversions][ijo_cnv]
    - [Edit > Options > Input/Output][ijo_i_o]
    - [Process > Binary > Options][ijo_bin]

    [ijo_app]: https://imagej.net/ij/docs/guide/146-27.html#sub:Appearance...
    [ijo_cnv]: https://imagej.net/ij/docs/guide/146-27.html#sub:Conversions...
    [ijo_col]: https://imagej.net/ij/docs/guide/146-27.html#sub:Colors...
    [ijo_i_o]: https://imagej.net/ij/docs/guide/146-27.html#sub:Input/Output...
    [ijo_bin]: https://imagej.net/ij/docs/guide/146-29.html#sub:BinaryOptions...
    """

    # Set all appearance settings to default values (untick all options)
    IJ.run("Appearance...", " ")

    # Set foreground color to be white and background black
    IJ.run("Colors...", "foreground=white background=black selection=red")

    # Options regarding binary images:
    # - `black`: set background for binary images to be black.
    # - `pad`: enable padding of edges to prevent eroding from image edge.
    # - `iterations=1`: number of times erosion (dilation, opening, closing) is
    #   performed
    # - `count=1`: number of adjacent background pixels necessary before a pixel
    #   is removed from the edge of an object during erosion and the number of
    #   adjacent foreground pixels necessary before a pixel is added to the edge
    #   of an object during dilation.
    # https://imagej.net/ij/docs/menus/process.html#options
    # https://imagej.net/ij/docs/guide/146-29.html#sub:BinaryOptions...
    IJ.run("Options...", "iterations=1 count=1 black pad")

    # Set default saving format to .txt files
    IJ.run("Input/Output...", "file=.txt save_column save_row")

    # Enable "scale when converting".
    # https://imagej.net/ij/docs/menus/edit.html#options
    IJ.run("Conversions...", "scale")



"""Helper functions to set up the scijava logger from Python."""

import sjlogging  # pylint: disable-msg=import-error
from .prefs import debug_mode


def scijava_logger(log_service):
    """Prepare logger and set the level according to stored ImageJ preferences.

    Parameters
    ----------
    log_service : org.scijava.log.LogService
        The LogService instance, usually retrieved in a SciJava script by using
        the script parameters annotation `#@ LogService logs` or equivalent.

    Returns
    -------
    logger : logging.Logger
        The Python logger object connected to SciJava's LogService.
    """
    logger = sjlogging.setup_logger(log_service)
    log_level = "INFO"
    if debug_mode():
        # issue a message with level "warn" to bring up the console window:
        logger.warn("Enabling debug logging.")
        log_level = "DEBUG"
    sjlogging.set_loglevel(log_level)
    return logger



"""String related helper functions."""

import re

from ._jython_compat import file_types


# this is taken from numpy's iotools:
def _is_string_like(obj):
    """Check whether obj behaves like a string.

    Using this way of checking for a string-like object is more robust when
    dealing with stuff that can behave like a 'str' but is not strictly an
    instance of it (or a subclass thereof). So it's more generic than using
    isinstance(obj, str).

    Parameters
    ----------
    obj : any
        The object to be checked for being string-like.

    Example
    -------
    >>> _is_string_like('foo')
    True
    >>> _is_string_like(123)
    False
    """
    try:
        obj + ""
    except (TypeError, ValueError):
        return False
    return True


def filename(name):
    """Get the filename from either a filehandle or a string.

    This is a convenience function to retrieve the filename as a string given
    either an open filehandle or just a plain str containing the name.

    When running in Jython the function will also convert `java.io.File` objects
    to `str`. NOTE: this also works if the Java object is a directory, not an
    actual file.

    Parameters
    ----------
    name : str or filehandle or java.io.File
        The object to retrieve the filename from.

    Returns
    -------
    name : str

    Example
    -------
    >>> filename('test_file_name')
    'test_file_name'

    >>> import os.path
    >>> fname = filename(open(__file__, 'r'))
    >>> os.path.basename(fname) in ['strtools.py', 'strtools.pyc']
    True
    """
    try:
        if isinstance(name, java.io.File):
            return str(name)
    except:
        # we silently ignore this case and continue with checks below as most
        # likely we are not running under Jython
        pass

    if isinstance(name, file_types):
        return name.name
    elif _is_string_like(name):
        return name
    else:
        raise TypeError


def flatten(lst):
    """Make a single string from a list of strings.

    Parameters
    ----------
    lst : list(str)
        The list of strings to be flattened.

    Returns
    -------
    flat : str

    Example
    -------
    >>> flatten(('foo', 'bar'))
    'foobar'
    """
    flat = ""
    for line in lst:
        flat += line
    return flat


def strip_prefix(string, prefix):
    """Remove a given prefix from a string.

    Parameters
    ----------
    string : str
        The original string from which the prefix should be removed.
    prefix : str
        The prefix to be removed.

    Returns
    -------
    str
        The original string without the given prefix. In case the original
        string doesn't start with the prefix, it is returned unchanged.
    """
    if string.startswith(prefix):
        string = string[len(prefix) :]
    return string


def sort_alphanumerically(data):
    """Sort a list alphanumerically.

    Parameters
    ----------
    data : list(str)
        List of strings to be sorted.

    Returns
    -------
    list

    Examples
    --------
    >>> sorted([ "foo-1", "foo-2", "foo-10" ])
    ["foo-1", "foo-10", "foo-2"]

    >>> sort_alphanumerically([ "foo-1", "foo-2", "foo-10" ])
    ["foo-1", "foo-2", "foo-10"]
    """

    def convert(text):
        return int(text) if text.isdigit() else text.lower()

    def alphanum_key(key):
        return [convert(c) for c in re.split("([0-9]+)", key)]

    return sorted(data, key=alphanum_key)


def pad_number(index, pad_length=2):
    """Pad a number with leading zeros to a specified length.

    Parameters
    ----------
    index : int or str
        The number to be padded
    pad_length : int, optional
        The total length of the resulting string after padding, by default 2

    Returns
    -------
    str
        The padded number as a string

    Examples
    --------
    >>> pad_number(7)
    '07'
    >>> pad_number(42, 4)
    '0042'
    """
    return str(index).zfill(pad_length)



"""ImageJ related functions, mostly convenience wrappers and combined workflows.

NOTE: this is only useful for Python (actually Jython) running within Fiji / ImageJ
and therefore will not be imported by the main 'imcflibs' package unless that
particular environment is detected.
"""

from . import bioformats
from . import misc
from . import prefs
from . import projections
from . import shading
from . import sjlog
from . import split
from . import stitching
